--[[
    This file creates a new Darktable module.
    copyright (c) 2022 Ulrich Gesing

    This script is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This script is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    
    See the GNU General Public License for more details
    at <http://www.gnu.org/licenses/>.

    This script can be used to do some configuration for an initial
    image workflow. It calls some automatisms of different modules
    in the darkroom view. If this suits your workflow, the script saves
    some clicks and time.
]]

--[[
    USAGE: Require this script from your luarc file, 
    add the path of this file to .config/darktable/luarc: 
    require "examples/InitialWorkflowModule"

    This script offers a new "inital workflow" module both in
    lighttable and darkroom view.

    Select one or more images and configure offered settings. Clicking the
    start button, selected image(s) are opened in darkroom. In the end,
    this script executes some automatic functions that can also be accessed
    via the GUI (magic wand).

    With Darktable 4.2, this script has been accelerated. It uses a new
    "pixelpipe-processing-complete event". It should be about twice as fast.

    The white balance is only adjusted by default for the modern workflow
    of Darktable. The script distinguishes between modern and legacy workflow
    and keeps white balance unchanged by default, if using legacy workflow.
    See Darktable preferences chapter "processing", setting
    auto-apply chromatic adaptation defaults.
    
    Offered steps:
    --Generate the shortest history stack that reproduces the current image.
    --Automatically adjust the exposure correction.
    --Enable and reset lens correction.
    --Adjust the white balance of the image by altering the temperature.
    --Perform color space corrections.
    --Enable and reset denoise profile.
    --Tone equalizer module: Automatic mask contrast + exposure compensation.
    --Color balance rgb module: Adjust global chroma and saturation.
    --Adjust filmic black + white relative exposure.
    --Discard the entire history stack and reactivate any default modules.

]]

local dt = require "darktable"
local du = require "lib/dtutils"
local log = require "lib/dtutils.log"
local debug = require "darktable.debug"

---------------------------------------------------------------

--log.log_level(log.warn)
--log.log_level(log.debug)
log.log_level(log.info)

-- check Darktable API version
-- new API of DT 4.2 is needed to use "pixelpipe-processing-complete" event
local minVersion = "9.0.0"
local minVersionInstalled = false
local status, err = pcall(function () du.check_min_api_version(minVersion, "InitialWorkflowModule") end)
if (status) then
  minVersionInstalled = true
  log.msg(log.info, "Darktable 4.2 API detected.")
end

-- get Darktable workflow setting
-- read preference "auto-apply chromatic adaptation defaults"
local function GetDarktableWorkflowSetting()
  return dt.preferences.read("darktable", "plugins/darkroom/chromatic-adaptation", "string")
end

-- debug helper function to dump preference keys
-- helps you to find out strings like "plugins/darkroom/chromatic-adaptation"
-- darktable -d lua > ~/keys.txt
local function DumpPreferenceKeys()
  log.msg(log.info, "get preference keys...")
  local keys = dt.preferences.get_keys()
  log.msg(log.info, #keys .. " retrieved, listing follows")
  for _,key in ipairs(keys) do
    log.msg(log.info, key)
  end
end

-- declare some variables to install the module
local env = {
  InstallModuleEventRegistered = false,
  InstallModuleDone = false,
  WaitForEventReceivedFlag = nil
}

-- collect gui controls
local GUI = {
  ProcessAllSelectedImagesButton = {},
  DisableAllGuiControlsButton = {},

  CompressHistoryStackCheckButton = {},
  ExposureCorrectionCheckButton = {},
  LensCorrectionCheckButton = {},
  WhiteBalanceComboBox = {},
  ColorCalibrationIlluminantComboBox = {},
  DenoiseProfileCheckButton = {},
  ToneEqualizerMaskCheckButton = {},
  ColorBalanceGlobalChromaComboBox = {},
  ColorBalanceGlobalSaturationComboBox = {},
  FilmicRelativeExposureCheckButton = {},
  ResetHistoryStackCheckButton = {}
}

local function DisableAllGuiControls()
  GUI.CompressHistoryStackCheckButton.value = false
  GUI.ExposureCorrectionCheckButton.value = false
  GUI.LensCorrectionCheckButton.value = false
  GUI.WhiteBalanceComboBox.value = 1
  GUI.ColorCalibrationIlluminantComboBox.value = 1
  GUI.DenoiseProfileCheckButton.value = false
  GUI.ToneEqualizerMaskCheckButton.value = false
  GUI.ColorBalanceGlobalChromaComboBox.value = 1
  GUI.ColorBalanceGlobalSaturationComboBox.value = 1
  GUI.FilmicRelativeExposureCheckButton.value = false
  GUI.ResetHistoryStackCheckButton.value = false
end

local function EnableDefaultGuiControls()
  GUI.CompressHistoryStackCheckButton.value = true
  GUI.ExposureCorrectionCheckButton.value = true
  GUI.LensCorrectionCheckButton.value = true
  
  -- distinguish between modern and legacy workflow
  -- keep white balance unchanged, if using legacy workflow
  -- see Darktable preferences - processing - auto-apply chromatic adaptation defaults
  if (GetDarktableWorkflowSetting() == "modern") then
    GUI.WhiteBalanceComboBox.value = 2
    GUI.ColorCalibrationIlluminantComboBox.value = 2
  else
    GUI.WhiteBalanceComboBox.value = 1
    GUI.ColorCalibrationIlluminantComboBox.value = 1
  end

  GUI.DenoiseProfileCheckButton.value = true
  GUI.ToneEqualizerMaskCheckButton.value = true
  GUI.ColorBalanceGlobalChromaComboBox.value = 5
  GUI.ColorBalanceGlobalSaturationComboBox.value = 7
  GUI.FilmicRelativeExposureCheckButton.value = true
  GUI.ResetHistoryStackCheckButton.value = false
end

local function AddGuiControlTooltips()
  GUI.ProcessAllSelectedImagesButton.tooltip = "Open selected image(s) in darkroom. Perform all configured steps for an initial workflow."
  GUI.DisableAllGuiControlsButton.tooltip = "Disable all steps of this module."
  
  GUI.CompressHistoryStackCheckButton.tooltip = "Same as compress history stack button: Generate the shortest history stack that reproduces the current image."
  GUI.ExposureCorrectionCheckButton.tooltip = "Exposure module: Automatically adjust the exposure correction."
  GUI.LensCorrectionCheckButton.tooltip = "Lens correction module: Enable and reset lens correction."
  
  GUI.WhiteBalanceComboBox.tooltip = "White balance module: Adjust the white balance of the image by altering the temperature. By default unchanged for the legacy workflow."
  GUI.ColorCalibrationIlluminantComboBox.tooltip = "Color calibration module: Perform color space corrections. By default unchanged for the legacy workflow."
  
  GUI.DenoiseProfileCheckButton.tooltip = "Denoise (profiled) module: Enable and reset denoise profile."
  GUI.ToneEqualizerMaskCheckButton.tooltip = "Tone equalizer module: Automatic mask contrast + exposure compensation."
  GUI.ColorBalanceGlobalChromaComboBox.tooltip = "Color balance rgb module: Adjust global chroma."
  GUI.ColorBalanceGlobalSaturationComboBox.tooltip = "Color balance rgb module: Adjust global saturation."
  GUI.FilmicRelativeExposureCheckButton.tooltip = "Filmic rgb module: Adjust filmic black + white relative exposure."
  GUI.ResetHistoryStackCheckButton.tooltip = "History: Discard the entire history stack and reactivate any default modules."

end

local function CreateGuiControls(processAllSelectedImagesFunction, disableAllGuiControls)
  -- create widget controls
  GUI.ProcessAllSelectedImagesButton = dt.new_widget("button"){label = "Start", clicked_callback = processAllSelectedImagesFunction}
  GUI.DisableAllGuiControlsButton = dt.new_widget("button"){label = "disable all", clicked_callback = disableAllGuiControls}
  
  GUI.CompressHistoryStackCheckButton = dt.new_widget("check_button"){label = "compress history stack"}
  GUI.ExposureCorrectionCheckButton = dt.new_widget("check_button"){label = "adjust exposure correction"}
  GUI.LensCorrectionCheckButton = dt.new_widget("check_button"){label = "enable lens correction"}
  
  GUI.WhiteBalanceComboBox = dt.new_widget("combobox"){label = "white balance", "unchanged", "as shot", "from image area"}
  GUI.ColorCalibrationIlluminantComboBox = dt.new_widget("combobox"){label = "color calibration illuminant", "unchanged"
  , "same as pipeline (D50)", "D (daylight)", "(AI) detect from image edges...", "as shot in camera"}
  
  GUI.DenoiseProfileCheckButton = dt.new_widget("check_button"){label = "reset denoise profile"}
  GUI.ToneEqualizerMaskCheckButton = dt.new_widget("check_button"){label = "tone equal. mask compensation"}
  GUI.ColorBalanceGlobalChromaComboBox = dt.new_widget("combobox"){label = "global chroma", "unchanged", 0,5,10,15,20,25,30,35}
  GUI.ColorBalanceGlobalSaturationComboBox = dt.new_widget("combobox"){label = "global saturation", "unchanged", 0,5,10,15,20,25,30,35}
  GUI.FilmicRelativeExposureCheckButton = dt.new_widget("check_button"){label = "filmic relative exposure"}
  GUI.ResetHistoryStackCheckButton = dt.new_widget("check_button"){label = "reset history stack"}

  AddGuiControlTooltips()
  EnableDefaultGuiControls()

end
 
---------------------------------------------------------------

local function WaitForEventReceived(event, image)      
  -- event handling
  -- log.msg(log.info, "received " .. event)
  env.WaitForEventReceivedFlag = 1
end 

local function WaitForEventRegister(eventType)
  -- event handling
  env.WaitForEventReceivedFlag = nil
  
  if (eventType) then
    dt.register_event("WorkflowModule", eventType, WaitForEventReceived)
  end
end

local function WaitForEvent(eventType, min, max)
  -- execute given function and wait for certain event type
  
  duration = 0
  period = 250 -- math.min(min,250)
  
  while (not (eventType and env.WaitForEventReceivedFlag)) or (duration < min) do
    
    dt.control.sleep(period)
    duration = duration + period

    if (duration >= max) then
      break
    end

  end  

  env.WaitForEventReceivedFlag = nil
  if (eventType) then
    dt.destroy_event("WorkflowModule", eventType)
  end
end

local function WaitForEventEmbeddedFunction(embeddedFunction, eventType, min, max)
  -- execute embedded function and wait for given event
  WaitForEventRegister(eventType)
  embeddedFunction()
  WaitForEvent(eventType, min, max)
end

local function WaitForPixelPipe(embeddedFunction)
  -- execute embedded function and wait for processing to completer

  if (minVersionInstalled) then
    -- dt 4.2: wait for new pixelpipe-processing-complete event
    WaitForEventRegister("pixelpipe-processing-complete")
    embeddedFunction()
    WaitForEvent("pixelpipe-processing-complete", 100, 2000)
  else
    -- old API
    WaitForEventRegister("darkroom-image-history-changed")
    embeddedFunction()
    WaitForEvent("darkroom-image-history-changed", 500, 2000)
  end
end

---------------------------------------------------------------

local function EnableDarkroomModule(moduleName)
  -- enable given darkroom module

  -- check if the module is already activated
  status = dt.gui.action(moduleName, 0, "enable", "", nil)
  if(status == 0) then
    log.msg(log.info, "enable module " .. moduleName)
    dt.gui.action(moduleName, 0, "enable", "", 1.0)
  end

  dt.gui.action(moduleName, 0, "show", "", 1.0)
end

---------------------------------------------------------------

local function AdjustExposureCorrection()  
  if (GUI.ExposureCorrectionCheckButton.value) then
    log.msg(log.info, "adjust the exposure correction")

    EnableDarkroomModule("iop/exposure")    

    WaitForPixelPipe(function()
      dt.gui.action("iop/exposure/exposure", 0, "button", "on", 1.0)
    end)    

    WaitForPixelPipe(function()
      dt.gui.action("iop/exposure/exposure", 0, "button", "off", 1.0)
    end)

  end
end

local function CompressHistoryStack()  
  if (GUI.CompressHistoryStackCheckButton.value) then
    log.msg(log.info, "compress the history stack")

    WaitForPixelPipe(function()
      dt.gui.action("lib/history/compress history stack", 0, "", "", 1.0)
    end)    

  end
end

local function AdjustToneEqualizerMask()  
  if (GUI.ToneEqualizerMaskCheckButton.value) then
    log.msg(log.info, "tone equalizer mask contrast + exposure compensation")

    EnableDarkroomModule("iop/toneequal")
    dt.gui.action("iop/toneequal/page", 0, "masking", "", 1.0)    

    WaitForPixelPipe(function()
      dt.gui.action("iop/toneequal/mask contrast compensation", 0, "button", "on", 1.0)
    end)

    WaitForPixelPipe(function()
      dt.gui.action("iop/toneequal/mask contrast compensation", 0, "button", "off", 1.0)
    end)

    WaitForPixelPipe(function()
      dt.gui.action("iop/toneequal/mask exposure compensation", 0, "button", "on", 1.0)
    end)

    WaitForPixelPipe(function()
      dt.gui.action("iop/toneequal/mask exposure compensation", 0, "button", "off", 1.0)
    end)

  end
end

local function AdjustColorBalance()
  if (GUI.ColorBalanceGlobalChromaComboBox.value ~= "unchanged") then
    log.msg(log.info, "set color balance global chroma")

    EnableDarkroomModule("iop/colorbalancergb")    

    WaitForPixelPipe(function()
      dt.gui.action("iop/colorbalancergb/global chroma", 0, "value", "set", GUI.ColorBalanceGlobalChromaComboBox.value / 100)
    end)    

  end

  if (GUI.ColorBalanceGlobalSaturationComboBox.value ~= "unchanged") then
    log.msg(log.info, "set color balance global saturation")

    EnableDarkroomModule("iop/colorbalancergb")    

    WaitForPixelPipe(function()
      dt.gui.action("iop/colorbalancergb/global saturation", 0, "value", "set", GUI.ColorBalanceGlobalSaturationComboBox.value / 100)
    end)    

  end

end

local function AdjustFilmicRelativeExposure()
  if (GUI.FilmicRelativeExposureCheckButton.value) then
    log.msg(log.info, "adjust filmic black + white relative exposure")

    EnableDarkroomModule("iop/filmicrgb")    

    WaitForPixelPipe(function()
      dt.gui.action("iop/filmicrgb/black relative exposure", 0, "button", "on", 1.0)
    end)

    WaitForPixelPipe(function()
      dt.gui.action("iop/filmicrgb/black relative exposure", 0, "button", "off", 1.0)
    end)

    WaitForPixelPipe(function()
      dt.gui.action("iop/filmicrgb/white relative exposure", 0, "button", "on", 1.0)
    end)

    WaitForPixelPipe(function()
      dt.gui.action("iop/filmicrgb/white relative exposure", 0, "button", "off", 1.0)
    end)

  end

end

local function ResetLensCorrection()
  if (GUI.LensCorrectionCheckButton.value) then
    log.msg(log.info, "reset lens correction")

    EnableDarkroomModule("iop/lens")    

    WaitForPixelPipe(function()
      dt.gui.action("iop/lens", 0, "reset", "", 1.0)
    end)    

  end
end

local function ResetHistoryStack(currentImage)
  if (GUI.ResetHistoryStackCheckButton.value) then
    log.msg(log.info, "reset history stack")    

    WaitForPixelPipe(function()
      dt.gui.action("lib/history", 0, "reset", "", 1.0)
    end)    
  end
end

local function ResetDenoiseProfile()
  if (GUI.DenoiseProfileCheckButton.value) then
    log.msg(log.info, "reset denoise profile")

    EnableDarkroomModule("iop/denoiseprofile")    

    WaitForPixelPipe(function()
      dt.gui.action("iop/denoiseprofile", 0, "reset", "", 1.0)
    end)    

  end
end

local function WhiteBalance(image)
  if (GUI.WhiteBalanceComboBox.value == "unchanged") then
    return
  end

  log.msg(log.info, "adapt white balance")

  EnableDarkroomModule("iop/temperature")

  if (GUI.WhiteBalanceComboBox.value == "as shot") then

    WaitForPixelPipe(function()
      dt.gui.action("iop/temperature/settings/as shot", 0, "", "", 1.0)
    end)    
    
  elseif (GUI.WhiteBalanceComboBox.value == "from image area") then

    WaitForPixelPipe(function()
      dt.gui.action("iop/temperature/settings/from image area", 0, "", "", 1.0)
    end)    
    
  end

end

local function AdjustColorCalibrationIlluminant(image)
  if (GUI.ColorCalibrationIlluminantComboBox.value == "unchanged") then
    return
  end

  log.msg(log.info, "adapt colorcalibration illuminant")

  EnableDarkroomModule("iop/channelmixerrgb")

  local newValue = GUI.ColorCalibrationIlluminantComboBox.value

  WaitForPixelPipe(function()
    dt.gui.action("iop/channelmixerrgb/illuminant", 0, "selection", "item:"..newValue, 1.0)
  end)    
  
 end

 local function ShowActiveModulesPage()
  log.msg(log.info, "show active modules page")
  dt.gui.action("lib/modulegroups/active modules", 0, "", "", 1.0)    
 end

local function LoadImage(image)
  -- load selected image and show it in lighttable view
  dt.gui.current_view(dt.gui.views.darkroom)  

  WaitForEventEmbeddedFunction(function()
    dt.gui.views.darkroom.display_image(image)
  end, "darkroom-image-loaded", 1000, 2000)  
end

local function GetTableLength(t)
  local counter = 0

  for n in pairs(t) do
    counter = counter + 1
  end

  return counter
end

local function ProcessImage(image)
  -- process one single image
  log.msg(log.info, "process image")

  --debug.max_depth = 3
  --log.msg(log.always, 4, dt.debug.dump(image))
  --local d = dt.gui.libs.modulegroups
  --log.msg(log.always, 4, dt.debug.dump(d))

  LoadImage(image)
  --ShowActiveModulesPage()
  ResetHistoryStack(image)
  WhiteBalance(image)
  AdjustColorCalibrationIlluminant()
  ResetDenoiseProfile(image)
  ResetLensCorrection()
  AdjustExposureCorrection()
  AdjustToneEqualizerMask()
  AdjustColorBalance()
  AdjustFilmicRelativeExposure()
  CompressHistoryStack()
  
  log.msg(log.info, "image processing done")
  log.msg(log.screen, "initial workflow - image processing done")
end

---------------------------------------------------------------

local function ProcessAllSelectedImages()
  -- function to process all selected images in lighttable view

  log.msg(log.info, "start button clicked")

  -- check that there is an image selected to activate darkroom view
  local selected_images = dt.gui.action_images
  if not selected_images or #selected_images == 0 then
    log.msg(log.screen, "no image selected")
    return
  end

  -- remember current view
  local current_view = dt.gui.current_view()

  -- process selected images
  for index, image in ipairs(selected_images) do 
    log.msg(log.info, "load image " .. index)
    
    ProcessImage(image)
  
  end

  -- return to lighttable view
  --dt.gui.current_view(current_view)

end
  
local function InstallModuleRegisterLib()
  -- register the module and create widged box in lighttable and darkroom

  if not env.InstallModuleDone then

    log.msg(log.info, "install module - create widget")
  
    dt.register_lib(
      "InitialWorkflowModule",    -- Module name
      "initial workflow",  -- name
      true,                -- expandable
      true,                -- resetable
      
      {[dt.gui.views.lighttable] = {"DT_UI_CONTAINER_PANEL_RIGHT_CENTER", 100},
      [dt.gui.views.darkroom] = {"DT_UI_CONTAINER_PANEL_LEFT_CENTER", 100}},
      
      dt.new_widget("box")
      {
        orientation = "vertical",
        reset_callback = function(this)
          EnableDefaultGuiControls()
        end,

        dt.new_widget("box")
        {
          orientation = "horizontal",
          
          GUI.ProcessAllSelectedImagesButton,
          GUI.DisableAllGuiControlsButton,
        },

        dt.new_widget("section_label"){label = "select workflow steps"},

        GUI.CompressHistoryStackCheckButton,
        GUI.FilmicRelativeExposureCheckButton,
        GUI.ColorBalanceGlobalChromaComboBox,
        GUI.ColorBalanceGlobalSaturationComboBox,
        GUI.ToneEqualizerMaskCheckButton,
        GUI.ExposureCorrectionCheckButton,
        GUI.LensCorrectionCheckButton,
        GUI.DenoiseProfileCheckButton,
        GUI.ColorCalibrationIlluminantComboBox,
        GUI.WhiteBalanceComboBox,

        dt.new_widget("section_label"){label = "reset all changes"},

        GUI.ResetHistoryStackCheckButton,
      },
      nil,-- view_enter
      nil -- view_leave
    )

    env.InstallModuleDone = true

  end
end

local function viewChangedEvent(event, old_view, new_view)
  -- register an event to signal changes from darkroom to lighttable
  log.msg(log.info, "view changed event")
  if new_view.name == "lighttable" and old_view.name == "darkroom" then
    InstallModuleRegisterLib()
  end
end

local function InstallModuleRegisterEvent()

  log.msg(log.info, "install module - register event")
  
  if not env.InstallModuleEventRegistered then
    dt.register_event("InitialWorkflowModule", "view-changed",viewChangedEvent)
    env.InstallModuleEventRegistered = true
  end
end

local function install_module()
  -- entry function to install the module at startup
  
  log.msg(log.info, "install module")
  
  CreateGuiControls(ProcessAllSelectedImages, DisableAllGuiControls)

  -- create the module depending on which view darktable starts in
  if dt.gui.current_view().id == "lighttable" then
    InstallModuleRegisterLib()
  else
    InstallModuleRegisterEvent()
  end

  return true
end

---------------------------------------------------------------

return install_module()

---------------------------------------------------------------
